
package shuteye;

import java.util.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.*;
import sleep.interfaces.*;
import sleep.runtime.*;
import sleep.bridges.BridgeUtilities;

import shuteye.classloader.*;

public class Classloader implements sleep.interfaces.Loadable
{
	private static Hashtable loaders = new Hashtable();
	
	public Classloader() {}

	//public boolean scriptLoaded(ScriptInstance script)
	public void scriptLoaded(ScriptInstance script)
	{
		loaders.put( script, new ExtendableClassLoader() );
		Hashtable environment = script.getScriptEnvironment().getEnvironment();
		environment.put("&addClassPath", new AddClassPath());
		environment.put("&getClassPath", new GetClassPath());
		environment.put("&loadClass", new LoadClass());
		environment.put("&useClass", new UseClass());
		environment.put("&useLoadable", new UseLoadable());
		
		AddLibPath libPath = new AddLibPath();
		environment.put("&addLibraryPath", libPath);
		environment.put("&addBinaryPath", libPath); // alias
		environment.put("&isLibraryPath", new IsLibPath());
		environment.put("&loadLibrary", new LoadLibrary());

		environment.put("&addSysClassPath", new AddSysClassPath());
		
		//return true;
	}

	//public boolean scriptUnloaded(ScriptInstance script)
	public void scriptUnloaded(ScriptInstance script)
	{
  		ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
  		if ( loader != null )
  			loader.removeAllPaths();
		loaders.remove( script );
  		loader = null;
		System.gc();
		//return true;
	}

	private static class AddClassPath implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			Scalar result = SleepUtils.getScalar( 0 );
			try
			{
				String path = BridgeUtilities.getString( args, "" );
				ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
				if ( loader.extendClasspath( path ) )
					result = SleepUtils.getScalar( 1 );
			}
			catch (Exception ex)
			{
				script.getScriptEnvironment().flagError(ex.toString());
			}
			return result;
		}
	}

	private static class GetClassPath implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			Scalar result = SleepUtils.getEmptyScalar();
			try
			{
				ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
				SleepUtils.getScalar( loader.getClasspath() );
			}
			catch (Exception ex)
			{
				script.getScriptEnvironment().flagError(ex.toString());
			}
			return result;
		}
	}
	
	private static class LoadClass implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			Scalar result = SleepUtils.getEmptyScalar();
			String classname = BridgeUtilities.getString( args, "" );
		//System.out.println(classname);
			try
			{
				ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
		//System.out.println(loader.toString());
				Class c = loader.findClass( classname );
		//System.out.println(c.toString());
				if ( c != null )
					result = SleepUtils.getScalar( c );
			}
			catch (Exception ex)
			{
				script.getScriptEnvironment().flagError(ex.toString());
			}
			return result;
		}
	}

	private static class UseClass implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			try
			{
				Object obj = BridgeUtilities.getObject( args );
				if ( obj instanceof Class && sleep.interfaces.Loadable.class.isAssignableFrom((Class)obj) )
				{
					Class bridge = (Class) obj;
					sleep.interfaces.Loadable temp = (sleep.interfaces.Loadable) bridge.newInstance();
					temp.scriptLoaded(script);
				}
				else if ( obj instanceof String  )
				{
					String bridge = (String) obj;
					ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
			//System.out.println(loader.toString());
					Class c = loader.findClass( bridge );
			//System.out.println(c.toString());
					if ( c != null && sleep.interfaces.Loadable.class.isAssignableFrom(c) )
					{
						sleep.interfaces.Loadable temp = (sleep.interfaces.Loadable) c.newInstance();
						temp.scriptLoaded(script);
						return SleepUtils.getScalar( 1 );
					}
				}
			}
			catch (Exception ex)
			{
				script.getScriptEnvironment().flagError(ex.toString());
			}
			return SleepUtils.getEmptyScalar();
		}
	}	

	private static class UseLoadable implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			sleep.interfaces.Loadable bridge = (sleep.interfaces.Loadable) BridgeUtilities.getObject( args );
			try
			{
				bridge.scriptLoaded(script);
			}
			catch (Exception ex)
			{
				script.getScriptEnvironment().flagError(ex.toString());
			}
			return SleepUtils.getEmptyScalar();
		}
	}	

	private static class AddLibPath implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			Scalar result = SleepUtils.getScalar( 0 );

			String path = BridgeUtilities.getString( args, "" );
			if ( path.equals("") ) return result;
			try
			{
				File libdir = new File(path).getCanonicalFile();
				if ( !libdir.isDirectory() )
					libdir = libdir.getParentFile();
				if ( !libdir.isDirectory() ) return result;
				Field usr_paths = ClassLoader.class.getDeclaredField("usr_paths");
				usr_paths.setAccessible(true);
				Object value = usr_paths.get(null);
				String[] currentpaths = (String[]) value;
				String[] newpaths = new String[currentpaths.length + 1];
				System.arraycopy( currentpaths, 0, newpaths, 0 , currentpaths.length );
				newpaths[currentpaths.length] = libdir.toString();
				value = newpaths;
				usr_paths.set(null,value);
				result = SleepUtils.getScalar( 1 );
			}
			catch ( Exception e )
			{
				script.getScriptEnvironment().flagError(e.toString());
			}
			return result;
		}
	}

	private static class IsLibPath implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			Scalar result = SleepUtils.getScalar( 0 );

			String path = BridgeUtilities.getString( args, "" );
			if ( path.equals("") ) return result;
			try
			{
				File libdir = new File(path).getCanonicalFile();
				path = libdir.toString();
				if ( !libdir.isDirectory() ) return result;
				Field usr_paths = ClassLoader.class.getDeclaredField("usr_paths");
				usr_paths.setAccessible(true);
				Object value = usr_paths.get(null);
				String[] currentpaths = (String[]) value;
				for ( int i=0; i<currentpaths.length; i++ )
				{
					if ( currentpaths[i].equals(path) ) 
					{
						result = SleepUtils.getScalar( 1 );
						break;
					}
				}
			}
			catch ( Exception e )
			{
				script.getScriptEnvironment().flagError(e.toString());
			}
			return result;
		}
	}

	private static class LoadLibrary implements Function
	{
		public Scalar evaluate(String name, ScriptInstance script, Stack args)
		{
			String library = BridgeUtilities.getString( args, "" );
			// ExtendableClassLoader loader = (ExtendableClassLoader) loaders.get( script );
			if ( library.equals("") ) return SleepUtils.getScalar(0);
			try
			{
				System.loadLibrary( library );
				return SleepUtils.getScalar(1);
			}
			catch ( Throwable e )
			{
				script.getScriptEnvironment().flagError(e.toString());
			}
			return SleepUtils.getScalar(0);
		}
	}
	
	/* Serge's code */
	
    private static final Class[] parameters = new Class[] { java.net.URL.class };

	
    private static class AddSysClassPath implements Function
    {

        public Scalar evaluate(String name, ScriptInstance script, Stack args)
        {
            Scalar result = SleepUtils.getScalar(0);
            try
            {
                String path = BridgeUtilities.getString(args, "");
                Classloader.addFile(path);
                result = SleepUtils.getScalar(1);
            }
            catch(Exception ex)
            {
                script.getScriptEnvironment().flagError(ex.toString());
            }
            return result;
        }
    }
    
    private static void addFile(String s) throws IOException
    {
        File f = new File(s);
        java.net.URL url = f.toURL();
        URLClassLoader sysloader = (URLClassLoader) ClassLoader.getSystemClassLoader();
        Class sysclass = java.net.URLClassLoader.class;
        try
        {
            Method method = sysclass.getDeclaredMethod( "addURL", parameters );
            method.setAccessible( true );
            method.invoke( sysloader, new Object[] { url } );
        }
        catch(Throwable t)
        {
            t.printStackTrace();
            throw new IOException("Error, could not add URL to system classloader");
        }
    }    
}
