
package shuteye.classloader;

import java.io.*;
import java.net.*;
import java.util.zip.*;

public class ExtendingClassLoader extends ClassLoader
{
	//ClassLoader parentloader = null;
	private String classpath = "";
	private boolean archive = false;
	private File parentdir = null;
	private ZipFile zipfile = null;
	
	private byte[] bytes = null;
	
	public ExtendingClassLoader(String path) throws InvalidClasspathException
	{
		super();
		checkPath( path );
	}
	
	public ExtendingClassLoader(ClassLoader father,String path) throws InvalidClasspathException
	{
		super(father);
		checkPath( path );
	}

	private void checkPath( String path ) throws InvalidClasspathException
	{
		classpath = getCanonicalPath( new File(path) );
		File file = new File( classpath );
		if (file.exists())
		{
			String lcPath = path.toLowerCase();
			if (file.isDirectory())
			{
				parentdir = file;
				archive = false;
			}
			else if (lcPath.endsWith(".jar") || lcPath.endsWith(".zip"))
			{
				try
				{
					zipfile = new ZipFile(file);
				}
				catch (IOException ioe)
				{
					throw new InvalidClasspathException( ioe.toString() );
				}
				archive = true;
			}
		}
		else throw new InvalidClasspathException( "file not found: " + classpath );
	}

	private String getCanonicalPath( File file ) throws InvalidClasspathException
	{
		String result = null;
		try
		{
			result = file.getCanonicalPath();
		}
		catch ( IOException ioe )
		{
			throw new InvalidClasspathException( ioe.toString() );
		}
		return result;
	}

	public Class findClass(String name)
	{
		byte[] b = null;
//		if (parentloader!=null)
//			b = parentloader.loadClassData(name);
//		if (b == null)	
//		{
//			b = loadClassData(name);
//		}
		b = loadClassData(name);
		if (b != null && b.length>0)
		{
			Class c_;
			try
			{
				c_ = defineClass(name, b, 0, b.length);
			}
			catch (NoClassDefFoundError nc) {last_error=""+nc;return null;}
			catch (LinkageError le) {last_error=""+le;return null;}
			catch (Exception e) {last_error=""+e;return null;}
			return c_;
		}
		else return null;
	}
	String last_error = "";
	public String getLastError()
	{
		String err = new String(last_error);
		last_error = "";
		return err;
	}
	
	public byte[] loadClassData(String classname)
 	{// load the class data from the connection
 		
 		if (archive)
 		{
// 			System.out.println("Archive: "+classpath);
 			return loadFromArchive(classAsFilename(classname,true));
 		}
 		else
 		{
// 			System.out.println("Directory: "+classpath);
 			return loadFromDirectory(classAsFilename(classname,false));
 		}
 	}
	
	public byte[] getBytecode()
	{
		return bytes;
	}
	
	public URL findResource(String resource)
	{
		URL u = null;
		try
		{
			if (archive)
			{
				String js = locateResource(resource);
				if (js!=null)// && classpath.toLowerCase().endsWith(".jar"))
				{
					u = new URL("jar:"+(new File(classpath).toURL())+"!/"+js);
				}
				else {}
			}
			else
			{
				String f = resourceAsFilename(resource);
				File r = new File(f);
				if (r.isAbsolute() && r.getPath().startsWith(classpath) && r.exists())
				{
					u = r.toURL();
				}
				else
				{
					File rs = new File(parentdir,f);
					if (rs.exists())
						u = rs.toURL();
				}
			}
		}
		catch (MalformedURLException me) {}
		return u;
	}
	
	public String locateResource(String resource)
	{
		if (archive)
		{
 				ZipEntry entry = zipfile.getEntry(resourceAsEntryname(resource));
 				if (entry!=null)
				{
 					//return classpath+"!"+resourceAsEntryname(resource);
					return resourceAsEntryname(resource);
				}
		}
		else
		{
			if (new File(parentdir,resourceAsFilename(resource)).exists())
			{
				//return classpath+File.separator+resourceAsFilename(resource);
				return resourceAsFilename(resource);
			}
		}
		return null;
	}
	
	public byte[] loadResource(String resource)
	{
		if (archive)
		{
			return loadFromArchive(resource);
		}
		else
		{
			return loadFromDirectory(resource);
		}
	}
	
	private byte[] loadFromArchive(String entryname)
	{
 		bytes = null;
 		if (zipfile!=null)
 		{
 			try
 			{
 				ZipEntry entry = zipfile.getEntry(entryname);
 				if (entry==null){return null;}
 				DataInputStream datastream = new DataInputStream(zipfile.getInputStream(entry));
				bytes = new byte[(int)entry.getSize()];
				datastream.readFully(bytes);
				datastream.close();
				return bytes;
 			}
 			catch (Exception e)
 			{
 				e.printStackTrace();
 				return null;
 			}
 		}
 		else {return null;}
  }
	
	private byte[] loadFromDirectory(String filename)
	{
		byte[] bytes = null;
		try
		{
			File classfile = new File(parentdir,filename);
//			System.out.println("classfile: "+classfile.toString());
			if (!classfile.exists()) {return null;}
			FileInputStream inputstream = new FileInputStream(classfile);
			DataInputStream datastream = new DataInputStream(inputstream);
			bytes = new byte[(int) classfile.length()];
			datastream.readFully(bytes);
			datastream.close();
			return bytes;
 		}
 		catch (Exception e)
 		{
 			e.printStackTrace();
 			return null;
 		}
	}
	
	private String classAsFilename(String classname,boolean zipped)
	{
		char sep = File.separatorChar;
		if (zipped) sep = '/';
		String result = classname.replace('.',sep)+".class";
//		System.out.println("filename: "+result);
		return result;
	}
	
	private String fileAsClassname(String classfilename)
	{
		String result = classfilename.replace('/','.');
		if (result.endsWith(".class"))
			result = result.substring(0,result.length()-6);
		return result;
	}
	
	private String resourceAsEntryname(String resource)
	{
		char sep = File.separatorChar;
		String result = resource;
		if (resource.indexOf("\\")!=-1)
			result = resource.replace('\\','/');
		
		return result;
	}
	
	private String resourceAsFilename(String resource)
	{
		char sep = File.separatorChar;
		String ssep = File.separator;
		String result = resource;
		if (ssep.equals("/") && (resource.indexOf("\\")!=-1))
			result = resource.replace('\\',sep);
		if (ssep.equals("\\") && (resource.indexOf("/")!=-1))
			result = resource.replace('/',sep);
		return result;
	}
	
	String getClasspath()
	{
		return classpath;
	}
	
	void closeFile()
	{
 		if ( zipfile != null )
 		{
 			try
 			{
 				zipfile.close();
 			}
 			catch ( Exception e ) { /* ignored */ }
 			finally { zipfile = null; }
 		}
	}
	
 	protected void finalize() throws Throwable
 	{
 		if ( zipfile != null )
 			closeFile();
 	}
	
}

